import logging
import math

from watermarks.core.utils import get_correct_wm
from watermarks.core.watermark import create_watermark
from watermarks.core.writers import BaseWriter

logger = logging.getLogger()


def update_parser(parser):
    pass


def init(args):
    '''Returns initialized Visible (writer) object from arguments passed from
    command line.
    '''
    wm = create_watermark(get_correct_wm(args, __name__.split('.')[-1]))
    return Visible(args.dest_dir, args.format, wm, args.suffix)


class Visible(BaseWriter):
    '''Visible method inserts visible watermark to top left corner of
    original image.

    '''
    allowed_formats = ('BMP', 'PNG', 'GIF', 'JPEG')
    allowed_modes = ('CMYK', 'L', 'RGB')

    def _create_watermarked(self, src_img):
        dst_img = src_img.copy()
        dst_w, dst_h = dst_img.size
        wm_w, wm_h = self.wm.img.size
        right = dst_w - wm_w
        bottom = dst_h - wm_h
        center_w = int(math.floor(right / 2.0) if right > 0 else
                       math.ceil(right / 2.0))
        center_h = int(math.floor(bottom / 2.0) if bottom > 0 else
                       math.ceil(bottom / 2.0))
        if self.position == 'TL':
            dst_img.paste(self.wm.img, (0, 0))
        elif self.position == 'T':
            dst_img.paste(self.wm.img, (center_w, 0))
        elif self.position == 'TR':
            dst_img.paste(self.wm.img, (right, 0))
        elif self.position == 'L':
            dst_img.paste(self.wm.img, (0, center_h))
        elif self.position == 'R':
            dst_img.paste(self.wm.img, (right, center_h))
        elif self.position == 'BL':
            dst_img.paste(self.wm.img, (0, bottom))
        elif self.position == 'B':
            dst_img.paste(self.wm.img, (center_w, bottom))
        elif self.position == 'BR':
            dst_img.paste(self.wm.img, (right, bottom))
        else:
            dst_img.paste(self.wm.img, (center_w, center_h))
        return dst_img



import logging
import os

from PIL import Image

from watermarks.core.readers import BaseReader


logger = logging.getLogger()


def update_parser(parser):
    pass


def init(args):
    '''Returns initialized Lsb (reader) object from arguments passed from
    command line.
    '''
    return Lsb(args.dest_dir, args.format, args.suffix,
               is_in_chain=args.is_in_chain)


class Lsb(BaseReader):
    '''Class wraps the LSB functionality. It allows to extract watermark
    from more images at once. To do so, just pass list of images (folders)
    to run() (argument `paths`). If the path is folder, it is scanned
    for images inside (not recursive). `destination` is path where extracted
    watermarks will be stored and `format` is their format (e.g. png).

    Class will generate watermark for each band separately. Generated
    watermark filepath is:
    "<destination>/<original filename>_<band shortcut>.<format>"
    '''
    allowed_formats = ('BMP', 'GIF', 'JPEG', 'PNG')
    allowed_modes = ('CMYK', 'L', 'RGB')

    def _create_watermarked(self, src_img):
        dst_imgs = []
        src_img.load()
        bands = src_img.split()
        for band_name, band in zip(src_img.getbands(), bands):
            bw = band.point(convert)
            dst_band = 'P' if src_img.mode == 'P' else 'L'
            dst_img = Image.merge(dst_band, (bw, ))
            dst_imgs.append(dst_img)
        return dst_imgs


def convert(x):
    '''Converts subpixel value to 0 or 255 depending on least significant
    bit.

    :param int x:
        Current subpixel value.
    :return:
        New subpixel value.
    :rtype: int
    '''
    return 255 if (x & 1) else 0



import os

from PIL import Image


def selftest(tmp='/tmp'):
    filepath = os.path.join(tmp, 'watermarks_test.png')
    _clean(filepath)
    try:
        _do_test(filepath)
        _clean(filepath)
        return True
    except Exception as edata:
        print('Error: %s' % edata)
        return False


def _do_test(filepath):
    sample_data = [1, 2, 3, 4]
    band = Image.new('L', (2, 2))
    band.putdata(sample_data)

    img = Image.merge('L', [band])
    img.save(filepath)

    if not os.path.exists(filepath):
        raise Exception('file was not created')

    img = Image.open(filepath)
    img.load()
    img_data = list(img.split()[0].getdata())

    if sample_data != img_data:
        raise Exception('image has invalid data')


def _clean(filepath):
    if os.path.exists(filepath):
        os.unlink(filepath)



import logging
import os
import math

import six
from PIL import Image


ALLOWED_FORMATS = ('BMP', 'JPEG', 'PNG')
# mode color depth, http://effbot.org/imagingbook/concepts.htm
MODE_DEPTHS = {
    '1': 1,
    'L': 8,
    'RGB': 8,
    'RGBA': 8,
    'CMYK': 8,
    # 'I': 32,
    # 'F': 32,
}
logger = logging.getLogger()


class Watermark(object):
    '''Basic class for image that will be put to other image as watermark.
    As input, it works with image object.
    '''
    def __init__(self, img):
        self.img = img

    def init(self):
        '''Initializes watermark - load, set width/height, ...'''
        if self.img.format not in ALLOWED_FORMATS:
            logger.warning('Watermark format "%s" is not allowed. (skip)',
                           self.img.format)
            raise ValueError()
        self.img.load()
        self.width, self.height = self.img.size
        self.band = self.img.split()[0]
        self.threshold = self.get_px_max_value() / 2

    def get_px_max_value(self):
        '''Determines maximum pixel value according to used mode.'''
        try:
            return 2 ** MODE_DEPTHS[self.img.mode] - 1
        except KeyError:
            logger.critical('Watermark mode "%s" is not supported!',
                            self.img.mode)
            raise


def create_watermark(wm, width=None, height=None, position=None, *args,
                     **kwargs):
    if isinstance(wm, six.string_types):
        if not os.path.isfile(wm):
            logger.critical('Watermark file "%s" does not exist!' % wm)
            exit(1)
        wm = Image.open(wm)

    if width and height:
        img_w, img_h = wm.size
        if img_w != width or img_h != height:
            right = width - img_w
            bottom = height - img_h
            center_w = int(math.floor(right / 2.0) if right > 0 else
                           math.ceil(right / 2.0))
            center_h = int(math.floor(bottom / 2.0) if bottom > 0 else
                           math.ceil(bottom / 2.0))
            sized_img = Image.new(wm.mode, (width, height), 'white')
            position = position.upper()
            if position == 'TL':
                sized_img.paste(wm, (0, 0))
            elif position == 'T':
                sized_img.paste(wm, (center_w, 0))
            elif position == 'TR':
                sized_img.paste(wm, (right, 0))
            elif position == 'L':
                sized_img.paste(wm, (0, center_h))
            elif position == 'R':
                sized_img.paste(wm, (right, center_h))
            elif position == 'BL':
                sized_img.paste(wm, (0, bottom))
            elif position == 'B':
                sized_img.paste(wm, (center_w, bottom))
            elif position == 'BR':
                sized_img.paste(wm, (right, bottom))
            else:
                sized_img.paste(wm, (center_w, center_h))
            sized_img.format = wm.format
            wm = sized_img

    wm_instance = Watermark(wm, *args, **kwargs)
    wm_instance.init()
    return wm_instance



import abc
import logging
import os

from PIL import Image

from watermarks.core.method import BaseMethod


logger = logging.getLogger()


class BaseWriter(BaseMethod):
    '''Base class for writers. It is responsible for selecting files to
    be processed, checking them and calling watermark method on them.

    '''
    __metaclass__ = abc.ABCMeta

    def __init__(self, destination, format_, wm, suffix='_watermarked',
                 position='C'):
        '''
        :param str destination:
            Destination where watermarked images will be stored.

        :param str format_:
            Output format.

        :param watermarks.core.watermark wm:
            Watermark instance.

        :param str suffix:
            Suffix added to generated files. If set to empty string,
            generated image will overwrite original image.

        :param str position:
            Watermark position. Allowed values are `TL`, `T`, `TR`, `L`,
            `C`, `R`, `BL`, `B`, `BR`.
        '''
        self.destination = destination
        self.format = format_
        self.wm = wm
        self.suffix = suffix
        self.position = position.upper()

    def _generate_files(self, filepath, src_img):
        base_name, _ = os.path.splitext(os.path.basename(filepath))
        logger.info('Processing file "%s"', filepath)
        dst_filepath = os.path.join(
            self.destination, '%s%s.%s' % (base_name, self.suffix, self.format)
        )
        dst_img = self._create_watermarked(src_img)
        dst_img.save(dst_filepath)
        logger.info('Generated file "%s".', dst_filepath)
        return [dst_filepath]



import abc
import logging
import os

from PIL import Image

from watermarks.core.method import BaseMethod


logger = logging.getLogger()


class BaseReader(BaseMethod):
    __metaclass__ = abc.ABCMeta

    def __init__(self, destination, format_, suffix='', is_in_chain=False):
        '''
        :param str destination:
            Destination where extracted watermarks will be stored.

        :param str format_:
            Watermark format.

        :param str suffix:
            Suffix added to generated files.

        :param bool is_in_chain:
            True if method is called in chain (more methods are applied).
        '''
        self.destination = destination
        self.format = format_
        self.suffix = suffix
        self.is_in_chain = is_in_chain

    def _generate_files(self, filepath, src_img):
        generated_filepaths = []
        base_name, _ = os.path.splitext(os.path.basename(filepath))
        logger.info('Processing file "%s"', filepath)
        dst_imgs = self._create_watermarked(src_img)
        class_name = self.__class__.__name__.lower()
        for band_name, dst_img in zip(src_img.getbands(), dst_imgs):
            dst_filepath = os.path.join(self.destination, '%s_%s%s%s.%s' % (
                base_name, band_name,
                '_%s' % class_name if self.is_in_chain else '',
                self.suffix, self.format))
            dst_img.save(dst_filepath)
            logger.info('Generated file "%s".', dst_filepath)
            generated_filepaths.append(dst_filepath)
        return generated_filepaths



import logging


def setup_logger(log_level=None):
    log_level = log_level or logging.INFO
    logger = logging.getLogger()
    logger.setLevel(log_level)
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)



import argparse
import logging
import sys

from watermarks import __version__ as version


class WMParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)


def handle_version():
    if '--version' in sys.argv:
        print(version)
        sys.exit(0)


def handle_help(parser):
    if '-h' in sys.argv or '--help' in sys.argv:
        parser.print_help()
        sys.exit(0)


def get_log_level(args):
    if args.verbose:
        ll = logging.DEBUG
    elif args.quiet:
        ll = logging.ERROR
    else:
        ll = None
    return ll


def add_basic_args(parser):
    vq_group = parser.add_mutually_exclusive_group()
    vq_group.add_argument(
        '-q', '--quiet', action='store_true', help='Be quiet.'
    )
    vq_group.add_argument(
        '-v', '--verbose', action='store_true', help='Be verbose.'
    )
    parser.add_argument(
        '--version', action='store_true', help='Show version and quit.'
    )
    parser.add_argument(
        '-h', '--help', action='store_true',
        help='Show this help message and exit.'
    )


def add_common_args(parser):
    parser.add_argument(
        'sources', metavar='PTH', nargs='+',
        help='List of files/directories that will be processed. Directories '
             'will be listed (but not recursive).'
    )
    parser.add_argument(
        '-d', '--dest-dir', required=True,
        help='Directory where the extracted watermarks will be stored.'
    )
    parser.add_argument(
        '--format', default='png',
        help='Format of generated images.'
    )
    parser.add_argument(
        '-s', '--suffix', default='',
        help='Suffix of watermarked files (default: %(default)s).'
    )



import logging


logger = logging.getLogger()


class Loader(object):
    '''Class responsible for loading and running watermark method.
    Method will first be searched in local modules and if it will not be
    found, Loader will try to find it in global scope (so you can use
    your own methods).
    '''
    def __init__(self, type_=''):
        '''
        :param str type_:
            readers/writers This parameter is used only for local methods
            and it is part of import path.
        '''
        self.type = type_
        self.modules = None

    def run(self, args, chaining=False):
        '''Runs desired watermark methods with `args`.'''
        new_files = []
        if not self.modules:
            self.load_methods(args.methods)
        for module in self.modules:
            x = module.init(args)
            if chaining:
                new_files = x.run(args.sources)
                args.sources = new_files
                args.suffix = ''
            else:
                new_files.extend(x.run(args.sources))
        return new_files

    def load_methods(self, methods):
        self.modules = []
        for method in methods.split(','):
            if not method:
                continue
            local_method = self.__module__.rsplit('.', 1)[0] + '.' \
                + self.type + '.' + method
            try:
                logger.debug('Trying to load method from local modules.')
                module = __import__(local_method,
                                    fromlist=('init', 'update_args'))
            except ImportError:
                logger.debug('Loading method from local modules failed, '
                             'trying to load from global scope.')
                module = __import__(method, fromlist=('init', 'update_args'))
            self.modules.append(module)



import abc
import logging
import os

from PIL import Image


logger = logging.getLogger()


class BaseMethod(object):

    def run(self, paths):
        '''Runs the process.

        :param list paths:
            Filepaths/folders to be processed.

        :return:
            List of generated files.
        :rtype: list
        '''
        processed = []
        if not os.path.exists(self.destination):
            os.makedirs(self.destination)
        for path in paths:
            if not os.path.exists(path):
                logger.error('Path "%s" does not exist! (skip)', path)
            elif os.path.isdir(path):
                processed.extend(self._process_dir(path))
            elif os.path.isfile(path):
                processed.extend(self._process_file(path))
            else:
                raise NotImplementedError('Cannot determine type of %s', path)
        return filter(None, processed)

    def _process_dir(self, dirpath):
        processed = []
        for filename in os.listdir(dirpath):
            full_filepath = os.path.join(dirpath, filename)
            if os.path.isfile(full_filepath):
                res = self._process_file(full_filepath)
                processed.extend(res)
        return processed

    def _process_file(self, filepath):
        src_img = Image.open(filepath)
        if src_img.format not in self.allowed_formats:
            logger.warning('File "%s" is in not allowed format. (skip)',
                           filepath)
            return []

        if src_img.mode in self.allowed_modes:
            return self._generate_files(filepath, src_img)
        else:
            logger.warning('File "%s" is in unsupported mode "%s". (skip)',
                           filepath, src_img.mode)
            return []

    @abc.abstractproperty
    def _create_watermarked(self, src_img):
        '''Method responsible for reading wm from single image.'''

    @abc.abstractproperty
    def allowed_formats(self):
        '''List of allowed image formats for particular method.'''

    @abc.abstractproperty
    def allowed_modes(self):
        '''List of allowed image modes for particular method.'''



import logging

from PIL import Image

from watermarks.core.utils import get_correct_wm
from watermarks.core.watermark import create_watermark
from watermarks.core.writers import BaseWriter

logger = logging.getLogger()


def update_parser(parser):
    pass


def init(args):
    '''Returns initialized Lsb (writer) object from arguments passed from
    command line.
    '''
    wm = create_watermark(get_correct_wm(args, __name__.split('.')[-1]))
    return Lsb(args.dest_dir, args.format, wm, args.suffix)


class Lsb(BaseWriter):
    '''Lsb (least significant bit) is method that changes least significant
    (last) bit for every subpixel in image according to reference image.
    With this approach you can generate new image almost identical to
    original image but with your hidden watermark.
    '''
    allowed_formats = ('BMP', 'PNG', 'GIF', 'JPEG')
    allowed_modes = ('CMYK', 'L', 'RGB')

    def _create_watermarked(self, src_img):
        src_img.load()
        bands = src_img.split()
        bands_wm = []
        src_width, src_height = src_img.size
        wm_width, wm_height = self.wm.img.size
        if src_width == wm_width and src_height == wm_height:
            wm = self.wm
        else:
            wm = create_watermark(self.wm.img, width=src_width,
                                  height=src_height, position=self.position)
        for band in bands:
            band_wm = Image.new('L', src_img.size)
            band_wm.putdata([convert(orig_px, wm_px, wm.threshold)
                             for orig_px, wm_px
                             in zip(band.getdata(), wm.band.getdata())
                             ])
            bands_wm.append(band_wm)
        dst_img = Image.merge(src_img.mode, bands_wm)
        return dst_img


def convert(orig_px, wm_px, threshold):
    '''Returns modified (last bit) value for subpixel.

    :param int orig_px:
        Current image subpixel value.
    :param int wm_px:
        Watermark subpixel value.
    :param int threshold:
        If `wm_px` is less or equal than this value, `orig_px` will
        change it's last bit value to 0 and to 1 if greater.
    :return:
        New subpixel value.
    :rtype: int
    '''
    if wm_px <= threshold:
        return orig_px & 254
    return orig_px | 1



from distutils.core import setup
from setuptools import find_packages

import os
import sys
ROOT_DIR = os.path.dirname(__file__)
SRC_DIR = os.path.join(ROOT_DIR, 'src')
sys.path.insert(0, SRC_DIR)
import watermarks


setup(
    name='watermarks',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    scripts=['bin/wm_reader', 'bin/wm_writer'],
    version=watermarks.__version__,
    description='Library for adding/reading watermarks from images. Currently visible and LSB watermark methods are supported.',
    author='Vladimir Chovanec',
    author_email='vladimir.chovanec.zc@gmail.com',
    maintainer_email='watermarks.py@gmail.com',
    keywords=['watermark', 'watermarks', 'watermarker', 'watermarking', 'lsb'],
    url='https://github.com/vladozc/watermarks',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Multimedia :: Graphics",
        ],
    install_requires=[
        'Pillow==2.4.0',
        'six==1.6.1',
        'argparse==1.2.1',
        ],
)
